@@ -8,9 +8,14 @@
    """
 
 import inspect
+from functools import partial
 import struct
 import re
-from .core import Message, FailReply
+import logging
+from .core import (Message, FailReply, DEFAULT_KATCP_MAJOR,
+                   SEC_TS_KATCP_MAJOR, SEC_TO_MS_FAC, MS_TO_SEC_FAC)
+
+logger = logging.getLogger(__name__)
 
 # KATCP Type Classes
 #
@@ -61,15 +66,16 @@ class KatcpType(object):
             raise ValueError("No value or default given")
         return self._default
 
-    def check(self, value):
+    def check(self, value, major):
         """Check whether the value is valid.
 
-        Do nothing if the value is valid. Raise an exception
-        if the value is not valid.
+        Do nothing if the value is valid. Raise an exception if the value is not
+        valid. Parameter major describes the KATCP major version to use when
+        interpreting the validity of a value.
         """
         pass
 
-    def pack(self, value, nocheck=False):
+    def pack(self, value, nocheck=False, major=DEFAULT_KATCP_MAJOR):
         """Return the value formatted as a KATCP parameter.
 
         Parameters
@@ -79,7 +85,8 @@ class KatcpType(object):
         nocheck : bool
             Whether to check that the value is valid before
             packing it.
-
+        major : int. Defaults to latest implemented KATCP version (5)
+             Major version of KATCP to use when interpreting types
         Returns
         -------
         packed_value : str
@@ -90,16 +97,18 @@ class KatcpType(object):
         if value is None:
             raise ValueError("Cannot pack a None value.")
         if not nocheck:
-            self.check(value)
-        return self.encode(value)
+            self.check(value, major)
+        return self.encode(value, major)
 
-    def unpack(self, packed_value):
+    def unpack(self, packed_value, major=DEFAULT_KATCP_MAJOR):
         """Parse a KATCP parameter into an object.
 
         Parameters
         ----------
         packed_value : str
             The unescaped KATCP string to parse into a value.
+        major : int. Defaults to latest implemented KATCP version (5)
+             Major version of KATCP to use when interpreting types
 
         Returns
         -------
@@ -109,9 +118,9 @@ class KatcpType(object):
         if packed_value is None:
             value = self.get_default()
         else:
-            value = self.decode(packed_value)
+            value = self.decode(packed_value, major)
         if value is not None:
-            self.check(value)
+            self.check(value, major)
         return value
 
 
@@ -128,9 +137,9 @@ class Int(KatcpType):
 
     name = "integer"
 
-    encode = lambda self, value: "%d" % (value,)
+    encode = lambda self, value, major: "%d" % (value,)
 
-    def decode(self, value):
+    def decode(self, value, major):
         try:
             return int(value)
         except:
@@ -141,7 +150,7 @@ class Int(KatcpType):
         self._min = min
         self._max = max
 
-    def check(self, value):
+    def check(self, value, major):
         """Check whether the value is between the minimum and maximum.
 
         Raise a ValueError if it is not.
@@ -166,9 +175,9 @@ class Float(KatcpType):
     """
 
     name = "float"
-    encode = lambda self, value: "%.15g" % (value,)
+    encode = lambda self, value, major: "%.15g" % (value,)
 
-    def decode(self, value):
+    def decode(self, value, major):
         try:
             return float(value)
         except:
@@ -179,7 +188,7 @@ class Float(KatcpType):
         self._min = min
         self._max = max
 
-    def check(self, value):
+    def check(self, value, major):
         """Check whether the value is between the minimum and maximum.
 
         Raise a ValueError if it is not.
@@ -197,9 +206,9 @@ class Bool(KatcpType):
 
     name = "boolean"
 
-    encode = lambda self, value: value and "1" or "0"
+    encode = lambda self, value, major: value and "1" or "0"
 
-    def decode(self, value):
+    def decode(self, value, major):
         if value not in ("0", "1"):
             raise ValueError("Boolean value must be 0 or 1.")
         return value == "1"
@@ -210,8 +219,8 @@ class Str(KatcpType):
 
     name = "string"
 
-    encode = lambda self, value: value
-    decode = lambda self, value: value
+    encode = lambda self, value, major: value
+    decode = lambda self, value, major: value
 
 
 class Discrete(Str):
@@ -236,7 +245,7 @@ class Discrete(Str):
             self._valid_values_lower = set([val.lower()
                                             for val in self._values])
 
-    def check(self, value):
+    def check(self, value, major):
         """Check whether the value in the set of allowed values.
 
         Raise a ValueError if it is not.
@@ -273,30 +282,81 @@ class Lru(KatcpType):
     ## @brief Mapping from LRU value name to LRU value constant.
     LRU_CONSTANTS = dict((v, k) for k, v in LRU_VALUES.items())
 
-    def encode(self, value):
+    def encode(self, value, major):
         if value not in Lru.LRU_VALUES:
             raise ValueError("Lru value must be LRU_NOMINAL or LRU_ERROR.")
         return Lru.LRU_VALUES[value]
 
-    def decode(self, value):
+    def decode(self, value, major):
         if value not in Lru.LRU_CONSTANTS:
             raise ValueError("Lru value must be 'nominal' or 'error'.")
         return Lru.LRU_CONSTANTS[value]
 
 
+class Address(KatcpType):
+    """The KATCP address type.
+
+    .. note::
+
+       The address type was added in katcp 0.4.
+    """
+
+    name = "address"
+
+    NULL = ("0.0.0.0", None)  # null address for use as an initial value
+
+    IPV4_RE = re.compile(r"^(?P<host>[^:]*)(:(?P<port>\d+))?$")
+    IPV6_RE = re.compile(r"^\[(?P<host>[^[]*)\](:(?P<port>\d+))?$")
+
+    def encode(self, value, major):
+        try:
+            host, port = value
+        except (ValueError, TypeError):
+            raise ValueError("Could not extract host and port from value %r" %
+                             (value,))
+        if ':' in host:
+            # IPv6
+            host = "[%s]" % host
+        return "%s:%s" % (host, port) if port is not None else host
+
+    def decode(self, value, major):
+        if value.startswith("["):
+            match = self.IPV6_RE.match(value)
+        else:
+            match = self.IPV4_RE.match(value)
+        if match is None:
+            raise ValueError("Could not parse '%s' as an address." % value)
+        port = match.group('port')
+        if port is not None:
+            port = int(port)
+        return match.group('host'), port
+
+
 class Timestamp(KatcpType):
     """The KATCP timestamp type."""
 
     name = "timestamp"
+    # Use microsecond precision, which is about as much as you can expect with a
+    # 64-bit float representing epoch-seconds
+    def encode(self, value, major):
+        if major >= SEC_TS_KATCP_MAJOR:
+            # In seconds, please!
+            return "%.6f" % float(value)
+        else:
+            # In milliseconds please!
+            return "%i" % int(float(value) * SEC_TO_MS_FAC)
 
-    encode = lambda self, value: "%i" % (int(float(value) * 1000),)
 
-    def decode(self, value):
+    def decode(self, value, major):
         try:
-            return float(value) / 1000
+            decoded = float(value)
         except:
             raise ValueError("Could not parse value '%s' as timestamp." %
                              value)
+        if major < SEC_TS_KATCP_MAJOR:
+            # Convert milliseconds to seconds
+            decoded = decoded * MS_TO_SEC_FAC
+        return decoded
 
 
 class TimestampOrNow(Timestamp):
@@ -312,40 +372,40 @@ class TimestampOrNow(Timestamp):
 
     NOW = object()
 
-    def encode(self, value):
+    def encode(self, value, major):
         if value is self.NOW:
             return "now"
-        return super(TimestampOrNow, self).encode(value)
+        return super(TimestampOrNow, self).encode(value, major)
 
-    def decode(self, value):
+    def decode(self, value, major):
         if value == "now":
             return self.NOW
-        return super(TimestampOrNow, self).decode(value)
+        return super(TimestampOrNow, self).decode(value, major)
 
 
 class StrictTimestamp(KatcpType):
-    """The a timestamp that enforces the XXXX.YYY format for timestamps.
+    """A timestamp that enforces the XXXX.YYY format for timestamps.
     """
 
     name = "strict_timestamp"
 
-    def encode(self, value):
+    def encode(self, value, major):
         try:
-            return "%.15g" % (value * 1000.0)
+            return "%.15g" % value
         except:
             raise ValueError("Could not encode value %r as strict timestamp." %
                              value)
 
-    def decode(self, value):
+    def decode(self, value, major):
         try:
             parts = value.split(".", 1)
             _int_parts = [int(x) for x in parts]
-            return float(value) / 1000.0
+            return float(value)
         except:
             raise ValueError("Could not parse value '%s' as strict timestamp."
                              % value)
 
-    def check(self, value):
+    def check(self, value, major):
         """Check whether the value is positive.
 
         Raise a ValueError if it is not.
@@ -370,14 +430,14 @@ class Struct(KatcpType):
         super(Struct, self).__init__(**kwargs)
         self._fmt = fmt
 
-    def encode(self, value):
+    def encode(self, value, major):
         try:
             return struct.pack(self._fmt, *value)
         except struct.error, e:
             raise ValueError("Could not pack %s into struct with format "
                              "%s: %s" % (value, self._fmt, e))
 
-    def decode(self, value):
+    def decode(self, value, major):
         try:
             return struct.unpack(self._fmt, value)
         except struct.error, e:
@@ -412,7 +472,7 @@ class Regex(Str):
             if self._compiled.flags & value])
         super(Regex, self).__init__(**kwargs)
 
-    def check(self, value):
+    def check(self, value, major):
         if not self._compiled.match(value):
             raise ValueError("Value '%s' does not match regex '%s' with flags"
                              " '%s'." % (value, self._pattern, self._flags))
@@ -435,10 +495,10 @@ class DiscreteMulti(Discrete):
 
     name = "discretemulti"
 
-    def encode(self, value):
+    def encode(self, value, major):
         return self.separator.join(sorted(value, key=str.lower))
 
-    def decode(self, value):
+    def decode(self, value, major):
         if self.all_keyword and value == self.all_keyword:
             return sorted(list(self._valid_values), key=str.lower)
         return sorted([v.strip() for v in value.split(self.separator)],
@@ -449,10 +509,10 @@ class DiscreteMulti(Discrete):
         self.separator = separator
         super(DiscreteMulti, self).__init__(values, **kwargs)
 
-    def check(self, value):
+    def check(self, value, major):
         """Check that each item in the value list is in the allowed set."""
         for v in value:
-            super(DiscreteMulti, self).check(v)
+            super(DiscreteMulti, self).check(v, major)
 
 
 class Parameter(object):
@@ -466,12 +526,15 @@ class Parameter(object):
         The parameter's name (introspected)
     kattype : KatcpType object
         The parameter's kattype
+    major : integer
+        Major version of KATCP to use when interpreting types
     """
 
-    def __init__(self, position, name, kattype):
+    def __init__(self, position, name, kattype, major):
         self.position = position
         self.name = name
         self._kattype = kattype
+        self.major = major
 
     def pack(self, value):
         """Pack the parameter using its kattype.
@@ -486,7 +549,7 @@ class Parameter(object):
         packed_value : str
             The unescaped KATCP string representing the value.
         """
-        return self._kattype.pack(value)
+        return self._kattype.pack(value, self.major)
 
     def unpack(self, value):
         """Unpack the parameter using its kattype.
@@ -503,7 +566,7 @@ class Parameter(object):
         """
         # Wrap errors in FailReplies with information identifying the parameter
         try:
-            return self._kattype.unpack(value)
+            return self._kattype.unpack(value, self.major)
         except ValueError, message:
             raise FailReply("Error in parameter %s (%s): %s" %
                             (self.position, self.name, message))
@@ -515,7 +578,7 @@ class Parameter(object):
 def request(*types, **options):
     """Decorator for request handler methods.
 
-    The method being decorated should take a sock argument followed
+    The method being decorated should take a req argument followed
     by arguments matching the list of types. The decorator will
     unpack the request message into the arguments.
 
@@ -525,16 +588,37 @@ def request(*types, **options):
         The types of the request message parameters (in order). A type
         with multiple=True has to be the last type.
 
+    Keyword Arguments
+    -----------------
+    include_msg : bool, default: False
+        Pass the request message as the third parameter to the decorated request
+        handler function
+    major : int. Defaults to latest implemented KATCP version (5)
+        Major version of KATCP to use when interpreting types
+
     Examples
     --------
     >>> class MyDevice(DeviceServer):
     ...     @request(Int(), Float(), Bool())
-    ...     @reply(Int(), Float())
-    ...     def request_myreq(self, sock, my_int, my_float, my_bool):
+    ...     @return_reply(Int(), Float())
+    ...     def request_myreq(self, req, my_int, my_float, my_bool):
     ...         return ("ok", my_int + 1, my_float / 2.0)
     ...
-    """
-    include_msg = options.get('include_msg', False)
+    ...     @request(Int(), include_msg=True)
+    ...     @return_reply(Bool())
+    ...     def request_is_odd(self, req, msg, my_int):
+    ...         self.inform(req, Message.reply_inform(
+    ...             msg, 'Checking oddity of %d' % my_int))
+    ...         return ("ok", my_int % 2)
+    ...
+
+"""
+    include_msg = options.pop('include_msg', False)
+    major = options.pop('major', DEFAULT_KATCP_MAJOR)
+    check_req = options.pop('_check_req', True)
+    if len(options) > 0:
+        raise TypeError('does not take keyword argument(s) %r.'
+                        % options.keys())
     # Check that only the last type has multiple=True
     if len(types) > 1:
         for type_ in types[:-1]:
@@ -553,11 +637,14 @@ def request(*types, **options):
             # We must be on the inside. Introspect the parameter names.
             all_argnames = inspect.getargspec(handler)[0]
 
-        # Slightly hacky way of determining whether there is a sock
-        has_sock = len(all_argnames) > 1 and all_argnames[1] == "sock"
+        # Slightly hacky way of determining whether there is a req
+        # For backward-compatibility also check for 'sock' and handle it the
+        # same as 'req'
+        has_req = check_req and len(all_argnames) > 1 and all_argnames[1] in (
+                "req", "sock")
 
         params_start = 1
-        if has_sock:
+        if has_req:
             params_start += 1
         if include_msg:
             params_start += 1
@@ -565,16 +652,16 @@ def request(*types, **options):
         argnames = all_argnames[params_start:]
 
         def raw_handler(self, *args):
-            if has_sock:
-                (sock, msg) = args
-                new_args = unpack_types(types, msg.arguments, argnames)
+            if has_req:
+                (req, msg) = args
+                new_args = unpack_types(types, msg.arguments, argnames, major)
                 if include_msg:
-                    return handler(self, sock, msg, *new_args)
+                    return handler(self, req, msg, *new_args)
                 else:
-                    return handler(self, sock, *new_args)
+                    return handler(self, req, *new_args)
             else:
                 (msg,) = args
-                new_args = unpack_types(types, msg.arguments, argnames)
+                new_args = unpack_types(types, msg.arguments, argnames, major)
                 if include_msg:
                     return handler(self, msg, *new_args)
                 else:
@@ -589,15 +676,50 @@ def request(*types, **options):
 
     return decorator
 
-inform = request
+# Using partial with no extra parameters lets us 'copy' the function so that we
+# can change the docstring without affecting the original function's docstring
+inform = partial(request)
 inform.__doc__ = """Decorator for inform handler methods.
 
        This is currently identical to the request decorator, and is
        thus an alias.
        """
 
+unpack_message = partial(request, _check_req=False)
+unpack_message.__doc__ = (
+"""Decorator that unpacks katcp.Messages to function arguments
 
-def return_reply(*types):
+The method being decorated should take arguments matching the list of types. The
+decorator will unpack the request message into the arguments.
+
+Parameters
+----------
+types : list of kattypes
+    The types of the request message parameters (in order). A type
+    with multiple=True has to be the last type.
+
+Keyword Arguments
+-----------------
+include_msg: bool, default: False
+    Pass the request message as the second parameter to the decorated request
+    handler function
+major : int. Defaults to latest implemented KATCP version (5)
+    Major version of KATCP to use when interpreting types
+
+Examples
+--------
+>>> class MyClient(DeviceClient):
+...     @unpack_message(Str(), Int(), Float(), Bool())
+...     def reply_myreq(self, status, my_int, my_float, my_bool):
+...         print 'myreq replied with ', (status, my_int, my_float, my_bool)
+...
+...     @unpack_message(Str(), Int(), include_msg=True)
+...     def inform_fruit_picked(self, msg, fruit, no_picked):
+...         print no_picked, 'of fruit ', fruit, ' picked.'
+...         print 'Raw inform message: ', str(msg)
+""")
+
+def return_reply(*types, **options):
     """Decorator for returning replies from request handler methods
 
     The method being decorated should return an iterable of result
@@ -613,15 +735,27 @@ def return_reply(*types):
     types : list of kattypes
         The types of the reply message parameters (in order).
 
+
+    Keyword Arguments
+    -----------------
+
+    major : int. Defaults to latest implemented KATCP version (5)
+        Major version of KATCP protocol to use when interpreting types
+
     Examples
     --------
     >>> class MyDevice(DeviceServer):
     ...     @request(Int())
-    ...     @reply(Int(), Float())
-    ...     def request_myreq(self, sock, my_int):
+    ...     @return_reply(Int(), Float())
+    ...     def request_myreq(self, req, my_int):
     ...         return ("ok", my_int + 1, my_int * 2.0)
     ...
     """
+    major = options.pop('major', DEFAULT_KATCP_MAJOR)
+    if len(options) > 0:
+        raise TypeError('return_reply does not take keyword argument(s) %r.'
+                        % options.keys())
+
     # Check that only the last type has multiple=True
     if len(types) > 1:
         for type_ in types[:-1]:
@@ -637,7 +771,7 @@ def return_reply(*types):
 
         def raw_handler(self, *args):
             reply_args = handler(self, *args)
-            return make_reply(msgname, types, reply_args)
+            return make_reply(msgname, types, reply_args, major)
         raw_handler.__name__ = handler.__name__
         raw_handler.__doc__ = handler.__doc__
 
@@ -652,16 +786,17 @@ def return_reply(*types):
     return decorator
 
 
-def send_reply(*types):
+def send_reply(*types, **options):
     """Decorator for sending replies from request callback methods
 
     This decorator constructs a reply from a list or tuple returned
     from a callback method, but unlike the return_reply decorator it
     also sends the reply rather than returning it.
 
-    The list/tuple returned from the callback method must have a sock
-    as its first parameter and the original message as the second. The
-    original message is needed to determine the message name and ID.
+    The list/tuple returned from the callback method must have req (a
+    ClientRequestConnection instance) as its first parameter and the original
+    message as the second. The original message is needed to determine the
+    message name and ID.
 
     The device with the callback method must have a reply method.
 
@@ -670,27 +805,38 @@ def send_reply(*types):
     types : list of kattypes
         The types of the reply message parameters (in order).
 
+    Keyword Arguments
+    -----------------
+
+    major : int. Defaults to latest implemented KATCP version (5)
+        Major version of KATCP to use when interpreting types
+
     Examples
     --------
     >>> class MyDevice(DeviceServer):
     ...     @send_reply('myreq', Int(), Float())
-    ...     def my_callback(self, msg, sock):
-    ...         return (sock, msg, "ok", 5, 2.0)
+    ...     def my_callback(self, req):
+    ...         return (req, "ok", 5, 2.0)
     ...
     """
+
+    major = options.pop('major', DEFAULT_KATCP_MAJOR)
+    if len(options) > 0:
+        raise TypeError('send_reply does not take keyword argument(s) %r.'
+                        % options.keys())
+
     def decorator(handler):
         def raw_handler(self, *args):
             reply_args = handler(self, *args)
-            sock = reply_args[0]
-            msg = reply_args[1]
-            reply = make_reply(msg.name, types, reply_args[2:])
-            self.reply(sock, reply, msg)
+            req = reply_args[0]
+            reply = make_reply(req.msg.name, types, reply_args[1:], major)
+            req.reply_with_message(reply)
         return raw_handler
 
     return decorator
 
 
-def make_reply(msgname, types, arguments):
+def make_reply(msgname, types, arguments, major):
     """Helper method for constructing a reply message from a list or tuple
 
     Parameters
@@ -701,16 +847,20 @@ def make_reply(msgname, types, arguments):
         The types of the reply message parameters (in order).
     arguments : list of objects
         The (unpacked) reply message parameters.
+    major : integer
+        Major version of KATCP to use when packing types
     """
     status = arguments[0]
     if status == "fail":
-        return Message.reply(msgname, *pack_types((Str(), Str()), arguments))
+        return Message.reply(
+            msgname, *pack_types((Str(), Str()), arguments, major))
     if status == "ok":
-        return Message.reply(msgname, *pack_types((Str(),) + types, arguments))
+        return Message.reply(
+            msgname, *pack_types((Str(),) + types, arguments, major))
     raise ValueError("First returned value must be 'ok' or 'fail'.")
 
 
-def unpack_types(types, args, argnames):
+def unpack_types(types, args, argnames, major):
     """Parse arguments according to types list.
 
     Parameters
@@ -721,6 +871,8 @@ def unpack_types(types, args, argnames):
         The arguments to parse.
     argnames : list of strings
         The names of the arguments.
+    major : integer
+        Major version of KATCP to use when packing types
     """
     if len(types) > 0:
         multiple = types[-1]._multiple
@@ -736,17 +888,17 @@ def unpack_types(types, args, argnames):
         name = ""
         if i < len(argnames):
             name = argnames[i]
-        params.append(Parameter(i+1, name, kattype))
+        params.append(Parameter(i+1, name, kattype, major))
 
 
     if len(args) > len(types) and multiple:
         for i in range(len(types), len(args)):
-            params.append(Parameter(i+1, name, kattype))
+            params.append(Parameter(i+1, name, kattype, major))
 
     # if len(args) < len(types) this passes in None for missing args
     return map(lambda param, arg: param.unpack(arg), params, args)
 
-def pack_types(types, args):
+def pack_types(types, args, major):
     """Pack arguments according the the types list.
 
     Parameters
@@ -755,6 +907,8 @@ def pack_types(types, args):
         The types of the arguments (in order).
     args : list of objects
         The arguments to format.
+    major : integer
+        Major version of KATCP to use when packing types
     """
     if len(types) > 0:
         multiple = types[-1]._multiple
@@ -766,13 +920,13 @@ def pack_types(types, args):
 
     if len(args) < len(types):
         # this passes in None for missing args
-        retvals = map(lambda ktype, arg: ktype.pack(arg), types, args)
+        retvals = map(lambda ktype, arg: ktype.pack(arg, major=major), types, args)
     else:
-        retvals = [ktype.pack(arg) for ktype, arg in zip(types, args)]
+        retvals = [ktype.pack(arg, major=major) for ktype, arg in zip(types, args)]
 
     if len(args) > len(types) and multiple:
         last_ktype = types[-1]
         for arg in args[len(types):]:
-            retvals.append(last_ktype.pack(arg))
+            retvals.append(last_ktype.pack(arg, major=major))
 
     return retvals
